#!/usr/bin/perl
# All Rights Reserved, Copyright (C) 2003, Hitachi Software Engineering Co., Ltd.
# $Id: domain_trans_lib.pl,v 1.1.1.1 2005/05/25 16:47:04 ynakam Exp $

require "common.pl";
use Fcntl;


#if domain is defined,new entry is created.
#key:domain name
#value:1
%Domain;


#Important structures for domain transition.
#key:current domain
#value:array of (parent domain,entry point)
%Domain_trans_rule;
#key:parent domain name
#value:array of (child domain,entry point)
%Domain_trans_tree;

#"user" sentence in middle language.
#key: domain name
#value: user names separated by comma.
%Users;

#Array of (fromdomain,program,todomain)
@Domain_trans_rule_array;

$Policy_root=set_policy_root();

&read_domain_trans();
&make_domain_trans_array(\@Domain_trans_rule_array);
&make_domain_trans_tree(\%Domain_trans_tree,\@Domain_trans_rule_array);


foreach (keys(%Domain_trans_tree)) {
    my $ref;
    $ref=$Domain_trans_tree{$_};
    @$ref=sort @$ref;
}


#$filename
sub set_domain_trans() {
    my $filename;
    my $domain;
    $filename=$_[0];

    #open(FILE,"<$filename") or die "File open error:$filename\n";
    sysopen(FILE,"$filename",O_RDONLY) or die "File open error:$filename\n";

    while (<FILE>)
    {
	my @s;

	#skip spaces in the head of line
	if (/[^\s\t]+.+/)
	{
	    $_=$&;
	}

	#skip comments
	if ( /[\s\t]+\#/||/^\#/)
	{
	    next;
	}

	if (/user[\s\t]/)
	{
	    my $user;
	    @s=split(/[\s\t]+/,$_);
	    $user=$s[1];
       
	    if ($user=~/;/)
	    {
		chop($user);
	    }
	    $Users{$domain}=$user;
	    next;
	}

	if (/domain[\s\t]/||/role[\s\t]/)
	{
	    @s=split(/[\s\t]+/,$_);
	    $domain=$s[1];
	    #strip ";" in the end of line
	    if ($domain=~/;/)
	    {
		chop($domain);
	    }
	    $Domain_trans_rule{$domain}=[];
	    $Domain{$domain}=1;
	    next;
	}

	if (/domain_trans/)
	{
	    my $from;
	    my $prog;
	    my $ref;
	    if ($domain eq "")
	    {
		print STDERR "Domain isn't defined in $filename";
		exit;
	    }
	    @s=split(/[\s\t]+/,$_);
	    $from=$s[1];
	    $prog=$s[2];
	    if ($prog=~/;/)
	    {
		chop($prog);
	    }
	    $ref=$Domain_trans_rule{$domain};
	    push @$ref,"$from,$prog";
	}
    }
  
    close(FILE);
}

#set value in  %Domain_trans
sub read_domain_trans() {
    opendir(DIR,$Policy_root) or die  "could not open $Policy_root";
    while ($_=readdir(DIR))
    {
	if (/\.a$/)
	{
	    my $filename;
	    $filename=$Policy_root."/".$_;
	    &set_domain_trans($filename);
	}
	else
	{
	    next;
	}
    }

    close(DIR);
}

#construct array from %Domain_trans_rule whose elements are (fromdomain,program,todomain) 
#argreference to an array 
sub make_domain_trans_array() {
    my $key;
    my $array_ref;
    $array_ref=$_[0];

    foreach $key (keys %Domain_trans_rule)
    {
	my $ref;
	my $tmp;
	$ref=$Domain_trans_rule{$key};

	foreach $tmp (@$ref)
	{
	    my $str;
	    $str=$tmp.",".$key;
	    push @$array_ref,$str;
	}
    }
}


#Construct %Domain_hash_tree from @Domain_trans_rule_array
#arg:1.reference to hash 2.reference to array
sub make_domain_trans_tree($$) {
    my $hash_ref;
    my $array_ref;

    $hash_ref=$_[0];
    $array_ref=$_[1];

    foreach (@$array_ref)
    {
	my @tmp;
	my $ref;
	@tmp=split(/,/);
	if (!exists $$hash_ref{$tmp[0]})
	{
	    $$hash_ref{$tmp[0]}=[];
	}

	$ref=$$hash_ref{$tmp[0]};
	push @$ref,"$tmp[2],$tmp[1]";
    }
}

#output domain sentence,role sentence,user sentence.
sub write_domain($$) {
    my $fh;
    my $domain;
    ($fh,$domain)=@_;

    if ($domain eq "global")
    {
	print $fh "domain global;\n";
	return;
    }
    if ($domain=~/_t$/)
    {
	print $fh "domain $domain;\n";
	return;
    }

    if ($domain=~/_r$/)
    {
	print $fh "role $domain;\n";
	my $users;
	my @tmp;
	$users=$Users{$domain};
	@tmp=split(/,/,$users);
	foreach (@tmp)
	{
	    print $fh "user $_;\n";
	}

	return;
    }

    die;
}

##########################################################
##Functions related to draw tree
##########################################################

#Hash to detect loop
#key:coordinate (x,y)
#value:domain name ,coordinate(x,y) of parent domain
%tree_node;

#if detect loop return 1
sub check_loop($$$) {
    my $x;
    my $y;
    my $to;

    ($to,$x,$y)=@_;

    while(1)
    {
	#If the name exists the same as "to" in the parent node,then loop.
	my $parent_x;
	my $parent_y;
	my $parent_domain;
	my @tmp;

	if (!exists($tree_node{"$x,$y"}))
	{
	    die;
	}

	@tmp=split(/,/,$tree_node{"$x,$y"});
	$parent_domain=$tmp[0];
	$parent_x=$tmp[1];
	$parent_y=$tmp[2];

	if ($parent_x<0)
	{
	    return 0;
	}

	if ($parent_domain eq $to)
	{
	    return 1;
	}

	$x=$parent_x;
	$y=$parent_y;
    }

    return 0;
}



sub check_exist($$$) {
    my $x;
    my $y;
    my $coord_ref;

    ($x,$y,$coord_ref)=@_;

    foreach (@$coord_ref)
    {
	my $tmpx;
	my $tmpy;
	($tmpx,$tmpy)=split(/,/,$_);
	if($y eq $tmpy && $x>$tmpx)
	{
	    return 1;
	}
    }

    return 0;
}

#check whether (x,from_y)-(x,to_y) crosses domain name. if it crosses ,return 1.
sub check_cross($$$$) {
    my $x;
    my $from_y;
    my $to_y;
    my $y;
    my $coord_ref;

    ($x,$from_y,$to_y,$coord_ref)=@_;

    for ($y=$from_y; $y<=$to_y; $y++)
    {
	if (&check_exist($x,$y,$coord_ref) == 1)
	{
	    return 1;
	}
    }

    return 0;
}

sub search_largest_y($$$) {
    my $x;
    my $y;
    my $largest=-1;
    my $coord_ref;

    ($x,$y,$coord_ref)=@_;

    foreach (@$coord_ref)
    {
	my $tmpx;
	my $tmpy;
	($tmpx,$tmpy)=split(/,/,$_);
#    print $tmpx,":",$tmpy,p;
	if ($x eq $tmpx && $tmpy>$y)
	{
	    if (&check_cross($x,$y,$tmpy,$coord_ref) == 0)
	    {
		$largest=$tmpy;
	    }
	}
    }

#  print $x,":",$y,":",$largest,p;

    return $largest;
}

sub draw_line($$$$$) {
    my $x;
    my $from_y;
    my $to_y;
    my $y;
    my $off_x;
    my $off_y;
    ($x,$from_y,$to_y,$off_x,$off_y)=@_;

    my $y;
    my $left_space=$x*24+10-5+$off_x;
    for ($y=$from_y;$y<=$to_y;$y++)
    {
	$top_space=$y*25+120+$off_y;
	print "<DIV STYLE=\"position:absolute;visibility:visible;left:$left_space;top:$top_space;font-size:12pt\" ID=\"t$y$y\">";
	print "<NOBR>";
	print "|";
	print "</NOBR>";
	print "</DIV>\n";
    }
}

#print "|"
#

sub print_line($$$) {
    my $dest;
    my $off_x;
    my $off_y;
    my $coord_ref;

    ($off_x,$off_y,$coord_ref)=@_;

    foreach (@$coord_ref)
    {
	my $x,$y;
	($x,$y)=split(/,/,$_);
	$dest=search_largest_y($x,$y,$coord_ref);
#    print "$x:$y:$dest",p;
	if ($y<=$dest)
	{
	    &draw_line($x,$y,$dest,$off_x,$off_y);
	}
    }
}


sub register_tree($) {
    my $ref;
    ($ref)=@_;
}


#Draw tree recursively based on %Domain_trans_tree
#$domain_name
sub print_tree($$$$$$$) {
    my $domain_name;
    my $x;
    my $y;
    my $parent_x;
    my $parent_y;
    my $coord_ref;#buffer for tree coordinate
	my $printed_node;#referrence to hash
	my $ref;

    ($domain_name,$x,$y,$parent_x,$parent_y,$coord_ref,$printed_node)=@_;

    my $left_space=$x*24+10;
    my $top_space=$y*25+120;

    push @$coord_ref,"$x,$y";

    my $parent_domain_name;
    my @tmp;
    @tmp=split(/,/,$tree_node{"$parent_x,$parent_y"});
    $parent_domain_name=$tmp[0];

    print "<DIV STYLE=\"position:absolute;visibility:visible;left:$left_space;top:$top_space;font-size:12pt\" ID=\"t$y\">";
    print "<NOBR>";

    print a({-href=>"Javascript:onClick=change(\'$domain_name\',\'$parent_domain_name\')"},"$domain_name");

    if (exists($$printed_node{$domain_name})&&(exists $Domain_trans_tree{$domain_name}))
    {
	print " ";
	print a({-href=>"Javascript:onClick=popup_domain_tree(\'$parent_domain_name\',\'$domain_name\')"},"---->");
    }
    print "</NOBR>";
    print "</DIV>\n";

    #if $domain_name is printed before print only node.
    if (exists($$printed_node{$domain_name}))
    {
	$y--;
	return $y;
    }
    $$printed_node{$domain_name}=1;

    $tree_node{"$x,$y"}="$domain_name,$parent_x,$parent_y";

    if (!exists($Domain_trans_tree{$domain_name}))
    {
	return $y-1;
    }

    $ref=$Domain_trans_tree{$domain_name};

    my %called_domain;
    my $current_x=$x;
    my $current_y=$y;
#    my $flag=0;

    foreach (@$ref)
    {
	my @tmp;
	my $d;
	@tmp=split(/,/);
	$d=$tmp[0];

	if (exists($called_domain{$d}))
	{				#This is not to display domains whose (from,to) is the same but prog is different.
	    next;
	}
	else
	{
	    $called_domain{$d}=1;
	}

	if (&check_loop($d,$current_x,$current_y) == 0)
	{				#detect loop
	    $y=&print_tree($d,$x+1,$y+1,$current_x,$current_y,$coord_ref,$printed_node);
	    $y++;
	}
	else
	{
	    # $left_space+=100;
	    $y++;
	    $x++;
	    my $left_space=$x*24+10;
	    my $top_space=$y*25+120;
	    push @$coord_ref,"$x,$y";
	    $x--;

	    print "<DIV STYLE=\"position:absolute;visibility:visible;left:$left_space;top:$top_space;font-size:12pt\" ID=\"t$y\">";
	    print "<NOBR>";

	#     if ($flag == 0)
# 	    {
# 		print a({-href=>"Javascript:onClick=change(\'$d\',\'$domain_name\')"},"$d")." ";
# 		print a({-href=>"Javascript:onClick=popup_domain_tree(\'$domain_name\',\'$d\')"},"---->");

# 		$flag=1;
# 	    }
# 	    else
# 	    {
	      
# 	      #print ",$d";
# 	    }
	    print a({-href=>"Javascript:onClick=change(\'$d\',\'$domain_name\')"},"$d")." ";
	    print a({-href=>"Javascript:onClick=popup_domain_tree(\'$domain_name\',\'$d\')"},"---->");

	    print "</NOBR>";
	    print "</DIV>\n";
	}
    }
    $y--;

    return $y;
}

1;
