#!/usr/bin/perl
# All Rights Reserved, Copyright (C) 2003, Hitachi Software Engineering Co., Ltd.
# $Id: selinux-lib.pl,v 1.1.1.1 2005/05/25 16:47:04 ynakam Exp $

###Functions used in File ACL configuration.
use Fcntl;
use FileHandle;
do "common.pl";


#File ACLs of global domain
#%Global_allow: key=path ,value:permission
#%Global_allowonly: key=path ,value:permission
%Global_allow;
%Global_allowonly;

########File ACLs of normal domain
#Domain_allow key=path ,value:permission
#Domain_allowonly
%Domain_allow;
%Domain_allowonly;

########hash to store "allow exclusive"
##key:label name
##value:permission
%Exc_allow;

#key:directory
#value:label names,separated by ","
%Exc_dir;

####registar filenames explicitly configured
#key:path ,value:1
%Explicit;
%Global_Explicit;

#path to configuration files
$Policy_root=set_policy_root();

read_global_file_acl();



sub print_file($) {
    my $filename;

    $filename=$_[0];

    #    open(FILE,"<$filename") or die "File open error:$filename\n";
    sysopen(FILE,"$filename",O_WRONLY|O_TRUNC| O_CREAT) or die "File open error:$filename\n";

    while (<FILE>)
    {
	print;
	print "<br>";
    }

    close(FILE);
}




sub set_file_acl_hash($***) {
    my $filename;
    local *allow_hash;
    local *allowonly_hash;
    local *explicit_hash;
    ($filename,*allow_hash,*allowonly_hash,*explicit_hash)=@_;

    #open(FILE,"<$filename") or die "File open error:$filename\n";
    sysopen(FILE,"$filename",O_RDONLY) or die "File open error:$filename\n";


    while (<FILE>)
    {
      #This section will be replaced by "&parse_allow_one_line" .
	my $path;
	my $permstr;
	my $perm;
	my $tmp;

	#skip spaces
	if (/[^\s\t]+.+/)
	{
	    $_=$&;
	}
	#skip comments
	if ( /[\s\t]+\#/||/^\#/)
	{
	    next;
	}

	#skip allowexclusive
	if (/[\s\t]exclusive[\s\t]/)
	{
	    next;
	}

	if (/^role[\s\t]/)
	{
	    next;
	}

	###analyze allow,allowonly,deny,denyonly
	if ( /^allow[\s\t]+/ )
	{
	    /\/\S*/;
	    $path=$&;
	    if ($path!~/\//)
	    {				#allow exclusive
		next;
	    }
	    chop_slash($path);
	    /[r,w,x,a,s]+[\s\t]*;/ ;
	    $permstr=$&;
	    $perm=str_to_perm($permstr);

	    $tmp=$allow_hash{$path};
	    if ($tmp ne "")
	    {
		$tmp|=$perm;
		$allow_hash{$path}=$tmp;
		$explicit_hash{$path}=1;
	    }
	    else
	    {
		$allow_hash{$path}=$perm;
		$explicit_hash{$path}=1;
	    }
	}
	elsif ( /^deny[\s\t]+/ )
	{
	    /\/[^(;\t\s)]*/;
	    $path=$&;
	    if ($path!~/\//)
	    {
		next;
	    }
	    chop_slash($path);
	    #   print "$path\n";
	    $allow_hash{$path}=$DENY;
	    $explicit_hash{$path}=1;
	}
	elsif (/[\s\t]*denyonly[\s\t]+/)
	{
	    /\/[^(;\t\s)]*/;
	    $path=$&;
	    chop_slash($path);
	    $allowonly_hash{$path}=$DENY;;
	    $explicit_hash{$path}=1;
	}
	elsif ( /^allowonly[\s\t]+/ )
	{
	    /\/\S*/;
	    $path=$&;
	    if ($path!~/\//)
	    {
		next;
	    }
	    chop_slash($path);
	    /[r,w,x,a,s]+[\s\t]*;/ ;
	    $permstr=$&;
	    $perm=str_to_perm($permstr);

	    $tmp=$allowonly_hash{$path};
	    if ($tmp ne "")
	    {
		$tmp|=$perm;
		$allowonly_hash{$path}=$tmp;
		$explicit_hash{$path}=1;
	    }
	    else
	    {
		$allowonly_hash{$path}=$perm;
		$explicit_hash{$path}=1;
	    }
	}
	elsif ( /^denyonly[\s\t]+/ )
	{
	    /\/[^(;\t\s)]*/;
	    $path=$&;
	    chop_slash($path);

	    $allowonly_hash{$path}=$DENY;
	}
	else
	{
	    #
	    if ( /{/||/}/ ||/domain/)
	    {
		;
	    }
	    else
	    {
		if (/\w/)
		{
		    ;
		}
	    }
	}
    }

    close(FILE);
}


##make %Domain_allow,%Domain_allowonly
sub read_domain_file_acl($) {
    my $domain=$_[0];
    my $filename;
    if ($domain eq "global")
    {
	print "ERROR"; die;
    }
    $filename=$Policy_root."/$domain".".a";
    &set_file_acl_hash($filename,Domain_allow,Domain_allowonly,Explicit);
}


##make %Global_allow,%Global_allowonly
sub read_global_file_acl()
{
    my $filename;

    $filename=$Policy_root."/global";

    &set_file_acl_hash($filename,Global_allow,Global_allowonly,Global_Explicit);
}


#make %Exc_dir from $filename
sub set_exc_dir($) {
    my $filename;
    my $dir;
    my $name;
    my $dummy;
    my $s;
    $filename=$_[0];
#  open(FILE,"<$filename") or die "File open error:$filename\n";
    sysopen(FILE,"$filename",O_RDONLY) or die "File open error:$filename\n";

    while (<FILE>)
    {
	#skip spaces
	if (/[^\s\t]+.+/)
	{
	    $_=$&;
	}

	#skip comments
	if ( /[\s\t]+\#/||/^\#/)
	{
	    next;
	}

	if (/exclusive/)
	{
	    @s=split(/[\s\t]+/,$_);
	    $dir=$s[1];
	    &chop_slash($dir);
	    $name=$s[3];
	    $name=~/[^;]*/;
	    $name=$&;
	    if (exists($Exc_dir{$dir}))
	    {
		if (!($Exc_dir{$dir}=~/$name/))
		{
		    $Exc_dir{$dir}=$Exc_dir{$dir}.",".$name;
		}
	    }
	    else
	    {
		$Exc_dir{$dir}=$name;
	    }
	}
	else
	{
	    next;
	}
    }

    close(FILE);
}

#make exc_allow
sub set_exc_allow($) {
    my $filename;
    my @s;
    my $dir;
    my $name;
    my $perm;
    my $permstr;
    $filename=$_[0];
#  open(INFILE,"<$filename") or die "File open error:$filename\n";
    sysopen(INFILE,"$filename",O_RDONLY) or die "File open error:$filename\n";

    while (<INFILE>)
    {
	#skip spaces
	if (/[^\s\t]+.+/)
	{
	    $_=$&;
	}

	#skip comments
	if ( /[\s\t]+\#/||/^\#/)
	{
	    next;
	}

	if (/exclusive/)
	{
	    @s=split(/[\s\t]+/,$_);
	    $dir=$s[1];
	    &chop_slash($dir);
	    $name=$s[3];
	    $name=~/[^;]*/;
	    $name=$&;
	    if (!exists($Exc_dir{$dir}))
	    {
		print "ERROR $dir";
		die;
	    }

	    if (exists($Exc_allow{$name}))
	    {
		$Exc_allow{$name}|=$EXC_CREATE;
	    }
	    else
	    {
		$Exc_allow{$name}=$EXC_CREATE;
	    }
	}

	if ( /[\s\t]*allow[\s\t]+/ )
	{
	    @s=split(/[\s\t]+/,$_);
	    if ($s[1]=~/\//)
	    {
		next;
	    }
	    $name=$s[1];
	    /[r,w,x,a,s]+[\s\t]*;/ ;
	    $permstr=$&;
	    $perm=str_to_perm($permstr);
	    if (exists($Exc_allow{$name}))
	    {
		$Exc_allow{$name}|=$perm;
	    }
	    else
	    {
		$Exc_allow{$name}=$perm;
	    }
	}
    }

    close(INFILE);
}


sub read_exc_file_acl($) {
    my $domain;
    my $filename;
    $domain=$_[0];

    #make %Exc_dir
    opendir(DIR,$Policy_root) or die  "could not open $Policy_root";
    while ($_=readdir(DIR))
    {
	if (/\.a$/)
	{
	    $filename=$Policy_root."/".$_;
	    &set_exc_dir($filename);
	}
	else
	{
	    next;
	}
    }
    close(DIR);

    #make %Exc_allow
    $filename=$Policy_root."/".$domain.".a";
    &set_exc_allow($filename);
}


#calculate permission of $domain(normal domain) to $path(directory)
sub calc_domain_default_perm($$) {
    my $perm;
    my $path;
    my $domain;

    ($path,$domain)=@_;
    $perm=$DENY;
    chop_slash($path) ;

    #When explicitly configured,return $perm
    $perm=$Domain_allow{$path};
    if ($perm ne "")
    {
	return $perm;
    }
    $perm=$Domain_allowonly{$path};
    if ($perm ne "")
    {
	return $perm;
    }
    $perm=$Global_allow{$path};
    if ($perm ne "")
    {
	return $perm;
    }
    $perm=$Global_allowonly{$path};
    if ($perm ne "")
    {
	return $perm;
    }


    ##when permission is not explicitly described ,permission is that of the nearest directory.
    while (1)
    {
	$perm=$Domain_allow{$path};
	if ($perm ne "")
	{
	    return $perm;
	}
	$perm=$Global_allow{$path};
	if ($perm ne "")
	{
	    if ($Domain_allowonly{$path} ne "")
	    {					#when allowonly is described,then permission is DENY
		return 0;
	    }
	    return $perm;
	}

	#go to parent directory
	if ( $path=~/\/.*\//)
	{
	    $path=$&;
	    chop_slash($path);
	}
	else
	{
	    last;
	}
    }

    $perm=$Domain_allow{'/'};
    if ($perm ne "")
    {
	return $perm;
    }

    $perm=$Global_allow{'/'};
    if ($perm ne "")
    {
	return $perm;
    }

    $perm=$DENY;

    return $perm;
}

#calculate permission of global domain to $path(directory)
sub calc_global_default_perm($$)
{
    my $perm;
    my $path;
    $perm=$DENY;
    $path=$_[0];

    chop_slash($path);


    $perm=$Global_allowonly{$path};
    if ($perm ne "")
    {
	return $perm;
    }

    #permission is that of the nearest parent directory.
    while (1)
    {
	$perm=$Global_allow{$path};
	if ($perm ne "")
	{
	    return $perm;
	}

	#go to parent directory
	if ( $path=~/\/.*\//)
	{
	    $path=$&;
	    chop_slash($path);
	}
	else
	{
	    last;
	}
    }

    $perm=$Global_allow{'/'};
    if ($perm ne "")
    {
	return $perm;
    }

    $perm=$DENY;

    return $perm;
}

#calc_perm(fullpath,domain)
#return permission of domain to fullpath(file,directory)
sub calc_perm($$) {
    my $fullpath;
    my $domain;
    my $path;
    my $perm;
    ($fullpath,$domain)=@_;

    #calculate directory where $fullpath belongs
    if ($fullpath=~/\/.*\//)
    {
	$path=$&;
    }
    else
    {
	$path="/";
    }

    if ( -d $fullpath)
    {
	$perm=calc_default_perm($fullpath,$domain);
    }
    else
    {
	#$perm=&calc_default_perm($path,$domain);
	if ($domain eq "global")
	{
	    $perm=$Global_allow{$fullpath};
	}
	else
	{
	    $perm=$Domain_allow{$fullpath};
	    if ($perm eq "")
	    {
		$perm=$Global_allow{$fullpath};
	    }
	}
	if ($perm eq "")
	{
	    $perm=calc_default_perm($path,$domain);
	}
    }

    return $perm;
}


#calc_default_perm(path,domain)
#calculate permission of $domain to $path(directory)
sub calc_default_perm($$) {
    my $perm;
    my $path;
    my $domain;
    ($path,$domain)=@_;
    $perm=$DENY;
    chop_slash($path) ;

    if ($domain eq "global")
    {
	$perm=&calc_global_default_perm($path,$domain);
    }
    else
    {
	$perm=&calc_domain_default_perm($path,$domain);
    }

    return $perm;
}

#output "allow,deny,allownly,denyonly"
#$fullpath
#$perm
#$onlyflag 1:allowonly
#$outfile
#$fh:file handle
sub print_allow($$$$) {
    my $fullpath;
    my $perm;
    my $onlyflag;
    my $permstr="";
    my $fh;


    ($fullpath,$perm,$onlyflag,$fh)=@_;

    if ($perm == $DENY)
    {
	if ($onlyflag == 1)
	{
	    print $fh "denyonly $fullpath;\n";
	}
	else
	{
	    print $fh "deny $fullpath;\n";
	}
    }
    else
    {
	if($onlyflag == 1)
	{
	    print $fh "allowonly $fullpath ";
	}
	else
	{
	    print $fh "allow $fullpath ";

	}

	if (($perm & $READ) != 0)
	{
	    $permstr.="r,";
	}

	if (($perm & $WRITE) != 0)
	{
	    $permstr.="w,";
	}

	if (($perm & $EXECUTE) != 0)
	{
	    $permstr.="x,";
	}

	if (($perm & $SEARCH) != 0)
	{
	    $permstr.="s,";
	}
	chop($permstr);		#chop ","

	print $fh "$permstr;\n";
    }
}


#$domain
#$fullpath
#$perm
#$onlyflag 1:allowonly
sub update_allow_hash($$$$) {
    my $domain;
    my $fullpath;
    my $perm;
    my $onlyflag;
    ($domain,$fullpath,$perm,$onlyflag)=@_;

    if ($domain eq "global")
    {
	if ($onlyflag==1)
	{
	    $Global_allowonly{$fullpath}=$perm;
	    delete $Global_allow{$fullpath};
	}
	else
	{
	    $Global_allow{$fullpath}=$perm;
	    delete $Global_allowonly{$fullpath};
	}
    }
    else
    {
	if ($onlyflag==1)
	{
	    $Domain_allowonly{$fullpath}=$perm;
	    delete $Domain_allow{$fullpath};
	}
	else
	{
	    $Domain_allow{$fullpath}=$perm;
	    delete $Domain_allowonly{$fullpath};

	}
    }
}



#$fh:filehande
#$domain:domain name
sub print_domain_set_file($$$) {
    my $fh;
    my $domain;
    my $dir;
    ($fh,$domain,$dir)=@_;
    &print_domain_file_acl($fh,$domain);
    &print_exc_file_acl($fh,$domain,$dir);
}


#if there is no "allow <dir> exclusive <exc_name>" then return 0
sub check_name($$) {
    my $exc_name;	#label of "allow exclusive"
    my  $perm;
    my $domain;
    %old_Exc_allow;
    %old_Exc_dir;

    $exc_name=$_[0];
    $domain=$_[1];
    opendir(DIR,$Policy_root) or die  "could not open $Policy_root";

    %old_Exc_allow=%Exc_allow;
    %old_Exc_dir=%Exc_dir;

    while ($_=readdir(DIR))
    {
	if (/\.a$/)
	{
	    $filename=$Policy_root."/".$_;
	    &set_exc_dir($filename);
	    &set_exc_allow($filename);
	    $perm=$Exc_allow{$exc_name};
	    if (($perm & $EXC_CREATE)!=0)
	    {
		close(DIR);
		%Exc_allow=%old_Exc_allow;
		%Exc_dir=%old_Exc_dir;
		return 1;
	    }
	}
	else
	{
	    next;
	}
    }

    close(DIR);
    %Exc_allow=%old_Exc_allow;
    %Exc_dir=%old_Exc_dir;

    return 0;
}


#return directory name where label of "allow exclusive" is defined.
sub get_exc_dir($) {
    my $name;
    $name=$_[0];

    my $dir;
    foreach $dir (keys %Exc_dir)
    {
	my @tmp;
	@tmp=split(/,/,$Exc_dir{$dir});

	foreach(@tmp)
	{
	    if ($_ eq $name)
	    {
		return $dir;
	    }
	}
    }

  return "";
}

sub print_exc_file_acl($$$) {
    my $fh;
    my $domain;
    my $current_dir;
    my $key;
    ($fh,$domain,$current_dir)=@_;

    foreach $key (keys %Exc_allow)
    {
	my $perm;
	my $permstr;
	#if ($key eq "allow"){
	#  die;
	#}

	$perm=$Exc_allow{$key};
	if ($perm==$DENY)
	{
	    next;
	}

	if (($perm & $EXC_CREATE) != 0)
	{
	    my $exc_dir;
	    $exc_dir=&get_exc_dir($key);
	    if ($exc_dir eq "")
	    {
		print STDERR "ERROR\n";
		die;
	    }
	    print $fh "allow $exc_dir exclusive $key;\n";
	    if ($perm == $EXC_CREATE)
	    {
		next;
	    }
	}

	#This is nessesary to delete entry when there is no "c"
	my $tmp;
	$tmp=&check_name($key,$domain);
	if ($tmp==0)
	{
	    next;
	}

	#write allow
	if (($perm & $READ) != 0)
	{
	    $permstr.="r,";
	}

	if (($perm & $WRITE) != 0)
	{
	    $permstr.="w,";
	}

	if (($perm & $EXECUTE) != 0)
	{
	    $permstr.="x,";
	}

	if (($perm & $SEARCH) != 0)
	{
	    $permstr.="s,";
	}
	chop($permstr);		#strip "," at the end of line

	print $fh "allow $key $permstr;\n";
    }
}

sub print_domain_file_acl($$) {
    my $fh;
    my $domain;
    my $key;
    ($fh,$domain)=@_;

    #print $fh "domain $domain;\n";
    #print_global_file_acl($fh);


    foreach $key (keys %Domain_allowonly)
    {
	my $perm;
	$perm=$Domain_allowonly{$key};

	&print_allow($key,$perm,1,$fh);
    }

    foreach $key (keys %Domain_allow)
    {
	my $perm;

	if (!exists($Domain_allow{$key}))
	{
	    next;
	}

	$perm=$Domain_allow{$key};
	&print_allow($key,$perm,0,$fh);
    }
}



sub print_global_file_acl($) {
    my $key;
    my $fh=$_[0];

    foreach $key (keys %Global_allowonly)
    {
	my $perm;
	$perm=$Global_allowonly{$key};
	&print_allow($key,$perm,1,$fh);
    }

    foreach $key (keys %Global_allow)
    {
	my $perm;
	$perm=$Global_allow{$key};
	&print_allow($key,$perm,0,$fh);
    }

    #close(OUTFILE);
}


sub print_global_set_file($) {
    my $fh=$_[0];
    print_global_file_acl($fh);
}


#make hash(key:label of allow exclusive,value:directory) from %Exc_dir
sub get_dir_exc() {
    my $exc_dir_hash_ref;
    my $new_hash_ref;
    ($exc_dir_hash_ref,$new_hash_ref)=@_;
    my $dir;

    foreach $dir (keys(%$exc_dir_hash_ref))
    {
	my @tmp;
	@tmp=split(/,/,$$exc_dir_hash_ref{$dir});
	my $key ;

	foreach $key (@tmp)
	{
	    $$new_hash_ref{$key}=$dir;
	}
    }
}


sub check_adm($$$) {
    my $domain;
    my $root;
    my $filename;
    ($domain,$root,$opt)=@_;

    if ($domain eq "global")
    {
	$filename=$root."/global";
    }
    else
    {
	$filename=$root."/$domain".".a";
    }
    sysopen(FILE,"$filename",O_RDONLY) or die "File open error:$filename\n";

    while (<FILE>)
    {
	my $line=$_;
	if ($line=~/^allowadm[\s\t]/&&$line=~/$opt/)
	{
	    close(FILE);
	    return 1;
	}
    }

    close(FILE);

    return 0;
}

sub check_search_all($$) {
    my $domain;
    my $root;
    my $filename;
    ($domain,$root)=@_;

    if ($domain eq "global")
    {
	$filename=$root."/global";
    }
    else
    {
	$filename=$root."/$domain".".a";
    }
    sysopen(FILE,"$filename",O_RDONLY) or die "File open error:$filename\n";

    while (<FILE>)
    {
	my $line=$_;
	if ($line=~/^allowadm[\s\t]/&&$line=~/search/)
	{
	    close(FILE);
	    return 1;
	}
    }

    close(FILE);

    return 0;
}


#insert configuration related to domain,file,perm,onlyflag
sub insert_permission($$){
  my $domain;
  my $file;
  my $perm;
  my $onlyflag;
  ($domain,$file,$perm,$onlyflag)=@_;

}


sub get_only_flag($){
  my $dir=$_[0];
  my $only_flag;
   if(exists($Domain_allowonly{$dir})){
    $only_flag=1;
  }elsif(exists($Global_allowonly{$dir})&&!exists($Domain_allow{$dir})){
    $only_flag=1;
  }else{
    $only_flag=0;
  }
  return $only_flag;
}


#update FileACL hashes
sub update_file_acl($$$$$$){
  my $domain;
  my $current_dir;
  my $default_perm;
  my $filename;
  my $perm;#permission to filename
  my $only_flag;
  my $parent_only_flag;

  ($domain,$current_dir,$default_perm,$filename,$perm,$only_flag,$parent_only_flag)=@_;

  local *allowhash;
  local *allowonlyhash;
  if ($domain eq "global"){
    *allowhash=*Global_allow;
    *allowonlyhash=*Global_allowonly;
  }else{
    *allowhash=*Domain_allow;
    *allowonlyhash=*Domain_allowonly;
  }

  my $fullpath;
  if ($filename eq "."||$filename eq ".."){
    return;
  }
  if ($current_dir eq "/"){
    $fullpath=$current_dir.$filename;
  }else{
    $fullpath=$current_dir."/".$filename;
  }

  #when $fullpath is directory
  if (-d $fullpath){
    if ($parent_only_flag==1){
      #allowonly
      if ($perm == $DENY){
	if (
	    #"deny" to $fullpath is described in advance.
	    exists($Domain_allow{$fullpath})||exists($Domain_allowonly{$fullpath})
	    #When "allow or deny or allowonly or denyonly" to $fullpath is described in global domain,
	    #if "deny" is not written,configuration in global domain will
	    #overwrite configuratin in normal domain.
	    ||exists($Global_allow{$fullpath})||exists($Global_allowonly{$fullpath})
	    #if denyonly is described to $fullpath,converter must write denyonly
	    ||$only_flag==1
	   ){
	  #tell converter to write "deny or denyonly"
	  &update_allow_hash($current_domain,$fullpath,$DENY,$only_flag);
	}else{
	  #in this case,access to $fullpath is implicitly denied.
	  #so converter shouldn't to write "deny".
	  delete $allowhash{$fullpath};
	  #delete $allowonlyhash{$fullpath};
	}
      }else{

	#when $perm is not DENY
	#if some of checkboxs of $fullpath is on,
	#converter should write "allow".Because permission isn't inherited from parent directory.
	&update_allow_hash($current_domain,$fullpath,$perm,$only_flag);
      }

    }else{
      #not allowonly
      if (
	  ####the conditions to write "allow,deny"
	  $perm!=$default_perm
	  ||$parent_only_flag!=$only_flag
	  #this condition is nessesary to overwrite configuration in global domain.
	  ||((exists($Global_allow{$fullpath})
	      &&($default_perm!=$Global_allow{$fullpath}))&&($current_domain ne "global"))
	  ||((exists($Global_allowonly{$fullpath})
	      &&($default_perm!=$Global_allowonly{$fullpath})&&($current_domain ne "global" ) ))
	 )
	{
	  #tell converter to write "allow or allowonly"
	  &update_allow_hash($current_domain,$fullpath,$perm,$only_flag);
	}else{
	  #in this case,access to $fullpath is implicitly allowed.
	  #so converter shouldn't to write "allow".
	  delete $allowhash{$fullpath};
	  delete $allowonlyhash{$fullpath};
	}
    }
  }else{
    #When $fullpath is file
        if ($perm!=$default_perm){
	  &update_allow_hash($current_domain,$fullpath,$perm,$only_flag);
	}elsif (((exists $Global_allow{$fullpath} ) &&
		 ($default_perm!=$Global_allow{$fullpath}) &&
		 ($current_domain ne "global"))  )
	  {
	    #this condition is nessesary to overwrite configuration in global domain.
	    &update_allow_hash($current_domain,$fullpath,$perm,$only_flag);
	    }else{
	      delete $allowhash{$fullpath};
	    }
      }

  #This is "deny" in global domain only
  #if ($perm==$DENY_EXPLICIT){
  if(($perm&$DENY_EXPLICIT)==$DENY_EXPLICIT){
    &update_allow_hash($current_domain,$fullpath,$DENY,$only_flag);
  }

  #don't write configuration the same as that of global domain.
  foreach $key (keys %Domain_allowonly){
    my $perm;
    $perm=$Domain_allowonly{$key};
    if (exists($Global_allowonly{$key})){
      if ($perm==$Global_allowonly{$key}){
	delete $Domain_allowonly{$key};
	next;
      }
    }
  }

  foreach $key(keys %Domain_allow){
    my $perm;
    $perm=$Domain_allow{$key};
    if (exists($Global_allow{$key})){
      if ($perm==$Global_allow{$key}){
	delete $Domain_allow{$key};
	next;
      }
    }
  }

}


sub reload_config($){
  my $domain=$_[0];
  undef %Global_allow;
  undef %Global_allowonly;
  undef %Domain_allow;
  undef %Domain_allowonly;
  undef %Exc_allow;
  undef %Exc_dir;
  undef %Explicit;
  undef %Global_Explicit;
  read_global_file_acl();
  if ($domain ne "global"){
    &read_domain_file_acl($domain);
    &read_exc_file_acl($domain);
  }
}

#update %Exc_allow
sub calc_exc_file_acl($) {
    my $key;
    my @name;
    my $current_dir=$_[0];

    if (!exists($Exc_dir{"$current_dir"}))
    {
	return;
    }

    @name=split(/,/,$Exc_dir{"$current_dir"});

    foreach (@name)
    {
	my $perm;
	if (!exists($permission{$_}))
	{
	    $perm=$DENY;
	}
	else
	{
	    $perm=$permission{$_};
	}
	$Exc_allow{$_}=$perm;
    }
}
